// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id: merl_video.c,v 1.5 2000/02/25 01:24:20 ersmith Exp $
//
// Copyright (C) 1993-1996 by id Software, Inc.
// Copyright (C) 1997 VM Labs, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log: merl_video.c,v $
// Revision 1.5  2000/02/25 01:24:20  ersmith
// version given to John for testing
//
// Revision 1.4  2000/02/11 17:35:15  ersmith
// updated for MiniBIOS and new BIOS routines
//
// Revision 1.3  2000/01/27 17:49:51  ersmith
// version linked to ROM BIOS
//
// Revision 1.2  2000/01/20 00:17:47  ersmith
// changed to use MPEAlloc to allocate extra MPEs; this allows multiple blitters to work right on MPE 3
//
// Revision 1.1.1.1  1999/12/20 14:23:42  ersmith
// Initial import of the VMLabs version of Doom
//
//
// DESCRIPTION:
//	DOOM graphics stuff for X11, UNIX.
//
//-----------------------------------------------------------------------------

//#define TEST_SPEED

#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <nuon/mml2d.h>
#include <nuon/mutil.h>
#include <nuon/joystick.h>
#include <nuon/bios.h>
#include <nuon/video.h>
#include <nuon/dma.h>

#include "doomstat.h"
#include "i_system.h"
#include "v_video.h"
#include "m_argv.h"
#include "d_main.h"

#include "doomdef.h"


#define X_DEADZONE	16
#define Y_DEADZONE	16


#ifdef WAIT_FOR_BLIT
// bitmask telling which blitters are busy
volatile unsigned long blit_busy;

#define MAX_MPES 32

static int
blit_comm_handler(int id, int comminfo, int res1, int res2, long p1, long p2, long p3, long p4)
{
    if (id >= MAX_MPES) {
	return id;
    }
    blit_busy &= ~(1U<<id);
    return -1;
}
#endif

/*
 * definitions for the MERLIN screen
 */

//#define DMA_WIDTH 344
//#define DMA_HEIGHT 208
#define DMA_WIDTH 360
#define DMA_HEIGHT 240

//
//  Translates button into a doom keycode
//  joydiff == joystick buttons that have changed
//  curjoy == current joystick value
//  but == joystick button to check
//  data == DOOM keycode equivalent
static inline void  
xlatekey(unsigned int joydiff, unsigned int curjoy,
	 unsigned int but, int data)
{
    event_t event;

    if (joydiff & but) {
	event.type = (curjoy & but) ? ev_keydown : ev_keyup;
	event.data1 = data;
	D_PostEvent(&event);
    }
}

void I_ShutdownGraphics(void)
{
}



//
// I_StartFrame
//
void I_StartFrame (void)
{
    // er?

}

void I_GetEvent(void)
{
    static unsigned long lastjoy;
    unsigned int curjoy;
    unsigned int joydiff;
    int xval, yval;
    event_t event;

    // put event-grabbing stuff in here

    curjoy = ((int)_Controller[1].buttonset|(int)_Controller[0].buttonset);
    joydiff = curjoy ^ lastjoy;
    if (joydiff) {
	// first, check for simulated "key" presses
	xlatekey(joydiff, curjoy, JOY_RIGHT, KEY_RIGHTARROW);
	xlatekey(joydiff, curjoy, JOY_LEFT, KEY_LEFTARROW);
	xlatekey(joydiff, curjoy, JOY_UP, KEY_UPARROW);
	xlatekey(joydiff, curjoy, JOY_DOWN, KEY_DOWNARROW);

	xlatekey(joydiff, curjoy, JOY_START, KEY_ESCAPE);
	xlatekey(joydiff, curjoy, JOY_Z, KEY_TAB);

	/* shoulder buttons */
	xlatekey(joydiff, curjoy, JOY_L, ' ');
	xlatekey(joydiff, curjoy, JOY_R, KEY_RALT);

	/* C buttons */
	xlatekey(joydiff, curjoy, JOY_C, KEY_MINUS); // Screen smaller
	xlatekey(joydiff, curjoy, JOY_D, '2'); // Cycle to weaker weapons
	xlatekey(joydiff, curjoy, JOY_E, '3'); // Cycle to stronger weapons
	xlatekey(joydiff, curjoy, JOY_F, KEY_EQUALS); // Screen bigger

	lastjoy = curjoy;
    }

    xval = _Controller[1].xAxis;
    yval = _Controller[1].yAxis;
    
    // Removed
    // if (xval < 0 && xval > -8) xval = 0;
    // if (yval < 0 && yval > -8) yval = 0;
    // xval = (xval >> 3) << 2;
    // yval = (yval >> 3) << 2;

    // Adjust values for deadzone
    if (xval > X_DEADZONE)
    	xval -= X_DEADZONE;
    else if (xval < -X_DEADZONE)
    	xval += X_DEADZONE;
    else
    	xval = 0;

    if (yval > Y_DEADZONE)
    	yval -= Y_DEADZONE;
    else if (yval < -Y_DEADZONE)
    	yval += Y_DEADZONE;
    else
    	yval = 0;
    	
    if (xval || yval || (joydiff & (JOY_A|JOY_B))) {
	// synthesize a "mouse" event
	event.type = ev_mouse;
	// the A and B buttons correspond to left and right mouse
	// buttons, respectively
	event.data2 = xval;
	event.data3 = yval;
	event.data1 = (curjoy & JOY_A) ? 1 : 0;
	event.data1 |= (curjoy & JOY_B) ? 4 : 0;
	D_PostEvent(&event);
    }

}

//
// I_StartTic
//
void I_StartTic (void)
{
    I_GetEvent();
}


//
// I_UpdateNoBlit
//
void I_UpdateNoBlit (void)
{
    // what is this?
}

#define NUM_DISPLAY_SCREENS 3
static mmlDisplayPixmap display_screen[NUM_DISPLAY_SCREENS];
static int renderScreen;  // which screen is being rendered to
static mmlGC gc;
static mmlSysResources sysRes;
static long	colors[256];   /* clut to use */
static scrInit[NUM_DISPLAY_SCREENS];

/* commands the DoScreenMsg function can understand */
#define SET_DEST_PARMS 1
#define SET_CLUT 2
#define SET_SOURCE_PARMS 3
#define DO_BLIT 4

static int mpenum[4];
static int lastmpe = 0;
static int use_blitter = 0;

#define MAX_HELPERS 4

/*
 * start up helper MPEs
 */

static void
StartHelpers(void)
{
    extern long blitd_start[], blitd_size[];
    extern long blitc_start[], blitc_size[];
    extern long mblitc_start[], mblitc_size[];

    int i;

    i = 0;

#ifdef USE_MPE0
    // first, try to get a MiniBios helper
    mpenum[0] = _MPEAlloc(MPE_HAS_MINI_BIOS);
    if (mpenum[0] >= 0) {
	i++;
	lastmpe++;
	StartMPE(mpenum[0], mblitc_start, (int)mblitc_size,
		 blitd_start, (int)blitd_size);
    }
#endif

    // now try allocating other helper MPEs
    for (; i < MAX_HELPERS; i++) {
	mpenum[i] = _MPEAlloc(0);
	printf("%d: mpe %d\n",i,mpenum[i]);
	if (mpenum[i] >= 0) {
	    StartMPE(mpenum[i],  blitc_start, (int)blitc_size,
		     blitd_start, (int)blitd_size);
	    lastmpe++;
	} else {
	    break;
	}
    }
    if (lastmpe == 0) {
	// couldn't allocate any helpers!!!
	_exit(0xdead0002);
    }
}

#ifndef INLINE
#define INLINE inline
#endif

//
// BlitWait -- wait for all blit processes to finish
//
void
BlitWait()
{
#ifdef WAIT_FOR_BLIT
    while (blit_busy != 0)
	;
#endif
}

#ifdef WAIT_FOR_BLIT
static INLINE void mark_busy(bits)
{
    asm __volatile__ ("st_io #1<<3,intctl"); // set software interrupt mask
    blit_busy |= bits;
    asm __volatile__ ("st_io #1<<2,intctl"); // clear software interrupt mask
}
#endif

/*
 * send a message to all blitter MPES
 */

static INLINE void
SendMsgToAll(long a, long b, long c, long d)
{
    int i;

    for (i = 0; i < lastmpe; i++) {
        /* send the request */
#ifdef WAIT_FOR_BLIT
	mark_busy(1<<mpenum[i]);
#endif
        _comm_send(a, b, c, d, mpenum[i], 0);
    }
}

/*
 * send a message to just one blitter MPE
 */

static INLINE void
SendMsgToOne(long a, long b, long c, long d)
{
    /* send the request */
#ifdef WAIT_FOR_BLIT
    mark_busy(1<<mpenum[use_blitter]);
#endif
    _comm_send(a, b, c, d, mpenum[use_blitter], 0);
}

static INLINE void
NextBlitter(void)
{
#ifdef WAIT_FOR_BLIT
    do {
	use_blitter++;
	if (use_blitter >= lastmpe)
	    use_blitter = 0;
    } while (0 != (blit_busy & (1<<mpenum[use_blitter])));
#else
    use_blitter++;
    if (use_blitter >= lastmpe)
	use_blitter = 0;
#endif
}

/*
 * set up the video parameters for future BlitToVideo calls
 */
static INLINE void
BlitSetup(mmlDisplayPixmap *dest)
{
    SendMsgToAll(SET_DEST_PARMS, (long)dest->memP, dest->dmaFlags, 0);
}

/*
 * set clut for blits
 */
static INLINE void
BlitSetClut(long *clut)
{
    _DCacheSync();
    SendMsgToAll(SET_CLUT, (long)clut, 0, 0);
}

/*
 * blit an 8bpp CLUT based image to a bilinear screen
 * "dest" is the destination pixmap
 * "image" is the base address of the 8bpp CLUT image
 * "imgwidth" is the width of the image
 * "destx" is the x position of the image in the destination
 * "desty" is the y position of the image in the destination
 * "srcx" is the x position of the image in the source
 * "srcy" is the y position of the image in the source
 * "width" is the width of the rectangle to copy
 * "height" is the height of the rectangle to copy
 *
 * we always add VIDEO_XOFF to destx and VIDEO_YOFF to desty
 */

#define VIDEO_XOFF ((DMA_WIDTH-SCREENWIDTH)/2)
#define VIDEO_YOFF ((DMA_HEIGHT-SCREENHEIGHT)/2)

void
BlitToVideo(unsigned char *image, int imgwidth,
	  int srcx, int srcy, int width, int height, int destx, int desty)
{
    if (width <= 0 || height <= 0)
        return;

    /* tell the video process about the image */
    SendMsgToOne(SET_SOURCE_PARMS, (long)image, 0, imgwidth);

    /* send a message to the video process to
       do the blit */
    SendMsgToOne(DO_BLIT, (width<<16)|(destx+VIDEO_XOFF), (height<<16)|(desty+VIDEO_YOFF),
                 0x00ffffff);

    SendMsgToOne(srcx<<16, srcy<<16, 0x00010000, 0);

    NextBlitter();
}

/*
 * draw a single column
 */
void
BlitColumn(unsigned char *image, int srcy, int srcystep,
	  int count, int destx, int desty, int shade)
{
    /* tell the video process about the image */
    SendMsgToOne(SET_SOURCE_PARMS, (long)image, 0, 128);

    /* send a message to the video process to
       do the blit */
    SendMsgToOne(DO_BLIT, (1<<16)|(destx+VIDEO_XOFF), (count<<16)|(desty+VIDEO_YOFF),
                 (shade<<16)|(127));

    SendMsgToOne(srcy, 0, 0, srcystep);

    NextBlitter();
}

/*
 * draw a texture span
 */
void
BlitSpan(unsigned char *image, int srcx, int srcy, int srcxstep, int srcystep,
	  int count, int destx, int desty, int shade)
{
    /* tell the video process about the image */
    SendMsgToOne(SET_SOURCE_PARMS, (long)image, 0, 64);

    /* send a message to the video process to
       do the blit */
    SendMsgToOne(DO_BLIT, (count<<16)|(destx+VIDEO_XOFF), (1<<16)|(desty+VIDEO_YOFF),
                 (shade<<16)|63);

    SendMsgToOne(srcx, srcy, srcxstep, srcystep);

    NextBlitter();
}

static INLINE void
FillWithColor(long flags, void *mem, int startx, int starty, int width, int height, long color)
{
    int x, y;

    for (y = starty; y < starty+height; y+=4) {
        for (x = startx; x < startx+width; x+=8) {
            _raw_plotpixel(flags, mem, (8<<16)|x, (4<<16)|y, color);
        }
    }
}

void
BlitClear(int cc, int x, int y, int width, int height)
{
    x += VIDEO_XOFF;
    y += VIDEO_YOFF;

    FillWithColor(display_screen[renderScreen].dmaFlags, display_screen[renderScreen].memP,
		   x, y, width, height, colors[cc]);
}

//
// I_FinishUpdate
//
void I_FinishUpdate (void)
{
    BlitWait();
//		waitkey();

#if 0
    /* swap the render screen and display screen */
    mmlSimpleVideoSetup(&display_screen[renderScreen], &sysRes, eTwoTapVideoFilter);
#else
    _VidSetup(display_screen[renderScreen].memP, display_screen[renderScreen].dmaFlags, DMA_WIDTH, DMA_HEIGHT, 2);
#endif
    renderScreen++;
    if (renderScreen >= NUM_DISPLAY_SCREENS)
        renderScreen = 0;

    if (!scrInit[renderScreen])
	{
		FillWithColor(display_screen[renderScreen].dmaFlags, display_screen[renderScreen].memP,
                  0, 0, DMA_WIDTH, DMA_HEIGHT, 0x10808000);
		scrInit[renderScreen]=1;
	}

    /* set up to render to the new screen */
    BlitSetup(&display_screen[renderScreen]);

//    _VidSync(0);
}


//
// I_ReadScreen
//
void I_ReadScreen (byte* scr)
{
#if 0
    memcpy (scr, screens[0], SCREENWIDTH*SCREENHEIGHT);
#endif
}


#define myFRACBITS 28
#define myFRACCONST (1<<myFRACBITS)
#define FRAC(f) ((long)((f)*myFRACCONST))

long
fastColorFromRGB(unsigned char r, unsigned char g, unsigned char b)
{
    int yf, crf, cbf;

    yf = FixMul(FRAC(0.299), r, myFRACBITS) +
	FixMul(FRAC(0.587), g, myFRACBITS) +
	FixMul(FRAC(0.114), b, myFRACBITS);
    yf = 16 + FixMul(FRAC(219.0/255.0), yf, myFRACBITS);
    if (yf < 16) yf = 16;

    crf = 128 + FixMul(FRAC(0.500), r, myFRACBITS) -
	FixMul(FRAC(0.419), g, myFRACBITS) -
	FixMul(FRAC(0.081), b, myFRACBITS);

    if (crf < 0) crf = 0;
    else if (crf > 255) crf = 255;

    cbf = 128 + FixMul(FRAC(0.500), b, myFRACBITS) -
	FixMul(FRAC(0.169), r, myFRACBITS) -
	FixMul(FRAC(0.331), g, myFRACBITS);

    if (cbf < 0) cbf = 0;
    else if (cbf > 255) cbf = 255;

    return (yf << 24) | (crf << 16) | (cbf << 8);

}

//
// I_SetPalette
//
void I_SetPalette (byte* palette)
{
    int i;
    int r,g,b;

    /* always set an alpha; it will be
       used for some translucent blits */
    for (i = 0; i < 256; i++) {
        r = gammatable[usegamma][*palette++];
        g = gammatable[usegamma][*palette++];
        b = gammatable[usegamma][*palette++];
#if 0
        colors[i] = mmlColorFromRGB(r, g, b) | 0xa0;
#else
	colors[i] = fastColorFromRGB(r, g, b) | 0xa0;
#endif
    }

    BlitSetClut(colors);
}


void I_InitGraphics(int what)
{
    int i;
	
	if (what&0x1)
	{
    	/* initialize the MML2D library */
    	mmlPowerUpGraphics(&sysRes);
    	mmlInitGC(&gc, &sysRes);

    	/* initialize the frame buffer */
    	/* use 32bit pixels for now */
    	mmlInitDisplayPixmaps( display_screen, &sysRes, DMA_WIDTH, DMA_HEIGHT, e888Alpha,
                        	   NUM_DISPLAY_SCREENS, NULL);

    	/* display the screen */
    	mmlSimpleVideoSetup(&display_screen[0], &sysRes, eTwoTapVideoFilter);
    	renderScreen = 1;


    	/* clear the screens to black */
    	for (i = 0; i < NUM_DISPLAY_SCREENS; i++) {
			scrInit[renderScreen]=0;
        	FillWithColor(display_screen[i].dmaFlags, display_screen[i].memP,
                    	  0, 0, DMA_WIDTH, DMA_HEIGHT, 0x10808000);
    	}
		DebugWS(display_screen[0].dmaFlags,display_screen[0].memP,50,100,0xA28E2C00,"Loading DOOM...");
	}
	if (what&0x02)	
    {
		StartHelpers();

    	/* set up for blitting to the "real" screen */

#ifdef WAIT_FOR_BLIT
    	/* initialize the routines for watching whether blitters are busy */
    	blit_busy = 0;
    	_IntSetVector(kIntrCommRecv, blit_comm_handler);
#endif

    	BlitSetup(&display_screen[renderScreen]);

    	/* clear the screens to black */
    	for (i = 0; i < NUM_DISPLAY_SCREENS; i++) {
        	FillWithColor(display_screen[i].dmaFlags, display_screen[i].memP,
                    	  0, 0, DMA_WIDTH, DMA_HEIGHT, 0x10808000);
			printf("screen %d @ %lx\n",i,display_screen[i].memP);
    	}
		DebugWS(display_screen[0].dmaFlags,display_screen[0].memP,50,100,0xA28E2C00,"Running DOOM...");
	}
}
